@@ -55,6 +55,9 @@
 #include "names.h"
 #include "devtree.h"
 
+#define _GNU_SOURCE
+#include <getopt.h>
+
 #define CTRL_RETRIES	 50
 #define CTRL_TIMEOUT	100	/* milliseconds */
 
@@ -64,9 +67,12 @@
 #define USB_DT_CS_INTERFACE		0x24
 #define USB_DT_CS_ENDPOINT		0x25
 
+#define VERBLEVEL_DEFAULT 0	/* 0 gives lspci behaviour; 1, lsusb-0.9 */
+
 static const char *procbususb = "/proc/bus/usb";
-static unsigned int verblevel = 1;
+static unsigned int verblevel = VERBLEVEL_DEFAULT;
 static int do_report_desc = 1;
+static int open_mode = O_RDONLY;
 
 static void dump_junk2(unsigned char *, unsigned int);
 
@@ -130,6 +136,7 @@ static int get_string(int fd, char *buf, size_t size, u_int8_t id, u_int16_t lan
 	b[0] = b[1] = 0xbf;
 	ret = usb_control_msg(fd, USB_DIR_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) | id, 0, sizeof(b), b);
 	if (ret < 0) {
+		if (open_mode == O_RDWR)
 		fprintf(stderr, "cannot get string descriptor %d, error = %s(%d)\n", id, strerror(errno), errno);
 		return 0;
 	}
@@ -867,6 +874,150 @@ static void dump_audiostreaming_endpoint(int fd, unsigned char *buf)
 	dump_junk(buf, "        ", 7);
 }
 
+static void dump_midistreaming_interface(int fd, unsigned char *buf, u_int16_t lang)
+{
+	static const char *jacktypes[] = {"Undefined", "Embedded", "External"};
+	char jackstr[128];
+	unsigned int j, tlength, capssize;
+	unsigned long caps;
+
+	if (buf[1] != USB_DT_CS_INTERFACE)
+		printf("      Warning: Invalid descriptor\n");
+	else if (buf[0] < 3)
+		printf("      Warning: Descriptor too short\n");
+	printf( "      MIDIStreaming Interface Descriptor:\n"
+		"        bLength             %5u\n"
+		"        bDescriptorType     %5u\n"
+		"        bDescriptorSubtype  %5u ",
+		buf[0], buf[1], buf[2]);
+	switch (buf[2]) {
+	case 0x01:
+		printf("(HEADER)\n");
+		if (buf[0] < 7)
+			printf("      Warning: Descriptor too short\n");
+		tlength = buf[5] | (buf[6] << 8);
+		printf( "        bcdADC              %2x.%02x\n"
+			"        wTotalLength        %5u\n",
+			buf[4], buf[3], tlength);
+		dump_junk(buf, "        ", 7);
+		break;
+
+	case 0x02:
+		printf("(MIDI_IN_JACK)\n");
+		if (buf[0] < 6)
+			printf("      Warning: Descriptor too short\n");
+		get_string(fd, jackstr, sizeof(jackstr), buf[5], lang);
+		printf( "        bJackType           %5u %s\n"
+			"        bJackID             %5u\n"
+			"        iJack               %5u %s\n",
+			buf[3], buf[3] < 3 ? jacktypes[buf[3]] : "Invalid",
+			buf[4], buf[5], jackstr);
+		dump_junk(buf, "        ", 6);
+		break;
+
+	case 0x03:
+		printf("(MIDI_OUT_JACK)\n");
+		if (buf[0] < 9)
+			printf("      Warning: Descriptor too short\n");
+		printf( "        bJackType           %5u %s\n"
+			"        bJackID             %5u\n"
+			"        bNrInputPins        %5u\n",
+			buf[3], buf[3] < 3 ? jacktypes[buf[3]] : "Invalid",
+			buf[4], buf[5]);
+		for (j=0; j < buf[5]; j++) {
+			printf( "        baSourceID(%2u)      %5u\n"
+				"        BaSourcePin(%2u)     %5u\n",
+				j, buf[2*j+6], j, buf[2*j+7]);
+		}
+		j = 6+buf[5]*2; /* midi10.pdf says, incorrectly: 5+2*p */
+		get_string(fd, jackstr, sizeof(jackstr), buf[j], lang);
+		printf( "        iJack               %5u %s\n",
+			buf[j], jackstr);
+		dump_junk(buf, "        ", j+1);
+		break;
+
+	case 0x04:
+		printf("(ELEMENT)\n");
+		if (buf[0] < 12)
+			printf("      Warning: Descriptor too short\n");
+		printf( "        bElementID          %5u\n"
+			"        bNrInputPins        %5u\n",
+			buf[3], buf[4]);
+		for(j=0; j < buf[4]; j++) {
+		printf( "        baSourceID(%2u)      %5u\n"
+			"        BaSourcePin(%2u)     %5u\n",
+			j, buf[2*j+5], j, buf[2*j+6]);
+		}
+		j = 5+buf[4]*2;
+		printf( "        bNrOutputPins       %5u\n"
+			"        bInTerminalLink     %5u\n"
+			"        bOutTerminalLink    %5u\n"
+			"        bElCapsSize         %5u\n",
+			buf[j], buf[j+1], buf[j+2], buf[j+3]);
+		capssize = buf[j+3];
+		caps = 0;
+		for(j=0; j < capssize; j++) {
+			caps |= (buf[j+9+buf[4]*2] << (8*j));
+		}
+		printf( "        bmElementCaps  0x%08lx\n", caps);
+		if (caps & 0x01)
+			printf( "          Undefined\n");
+		if (caps & 0x02)
+			printf( "          MIDI Clock\n");
+		if (caps & 0x04)
+			printf( "          MTC (MIDI Time Code)\n");
+		if (caps & 0x08)
+			printf( "          MMC (MIDI Machine Control)\n");
+		if (caps & 0x10)
+			printf( "          GM1 (General MIDI v.1)\n");
+		if (caps & 0x20)
+			printf( "          GM2 (General MIDI v.2)\n");
+		if (caps & 0x40)
+			printf( "          GS MIDI Extension\n");
+		if (caps & 0x80)
+			printf( "          XG MIDI Extension\n");
+		if (caps & 0x100)
+			printf( "          EFX\n");
+		if (caps & 0x200)
+			printf( "          MIDI Patch Bay\n");
+		if (caps & 0x400)
+			printf( "          DLS1 (Downloadable Sounds Level 1)\n");
+		if (caps & 0x800)
+			printf( "          DLS2 (Downloadable Sounds Level 2)\n");
+		j = 9+2*buf[4]+capssize;
+		get_string(fd, jackstr, sizeof(jackstr), buf[j], lang);
+		printf( "        iElement            %5u %s\n", buf[j], jackstr);
+		dump_junk(buf, "        ", j+1);
+		break;
+
+	default:
+		printf("\n        Invalid desc subtype: ");
+		dump_junk2(buf+3, buf[0]-3);
+		break;
+	}
+}
+
+static void dump_midistreaming_endpoint(int fd, unsigned char *buf)
+{
+	unsigned int j;
+
+	if (buf[1] != USB_DT_CS_ENDPOINT)
+		printf("      Warning: Invalid descriptor\n");
+	else if (buf[0] < 5)
+		printf("      Warning: Descriptor too short\n");
+	printf("        MIDIStreaming Endpoint Descriptor:\n"
+	       "          bLength             %5u\n"
+	       "          bDescriptorType     %5u\n"
+	       "          bDescriptorSubtype  %5u (%s)\n"
+	       "          bNumEmbMIDIJack     %5u\n",
+	       buf[0], buf[1], buf[2], buf[2] == 1 ? "GENERAL" : "Invalid", buf[3]);
+	for (j=0; j<buf[3]; j++) {
+		printf("          baAssocJackID(%2u)   %5u\n", j, buf[4+j]);
+	}
+	dump_junk(buf, "          ", 4+buf[3]);
+}
+
+
 static void dump_hub(char *p)
 {
 	unsigned int l, i, j;
@@ -1009,6 +1160,7 @@ static void dump_hid_device(int fd, unsigned char *buf,int interface_number)
 			continue;
 		}
 		if ((n = usb_control_msg(fd, 0x81 , 0x06, 0x22<<8, interface_number, len, dbuf)) < 0) {
+			if (open_mode == O_RDWR)
 			printf("cannot get report descriptor\n");
 			continue;
 		}
@@ -1027,6 +1179,7 @@ static void do_config(int fd, unsigned int nr, u_int16_t lang)
 
 	if (usb_control_msg(fd, USB_DIR_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_CONFIG << 8) | nr,
 			    0, USB_DT_CONFIG_SIZE, buf) < 0) {
+		if (open_mode == O_RDWR)
 		fprintf(stdout, "cannot get config descriptor %d, %s (%d)\n", nr, strerror(errno), errno);
 		return;
 	}
@@ -1039,6 +1192,7 @@ static void do_config(int fd, unsigned int nr, u_int16_t lang)
 	}
 	if (usb_control_msg(fd, USB_DIR_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_CONFIG << 8) | nr,
 			    0, sz, buf) < 0) {
+		if (open_mode == O_RDWR)
 		fprintf(stdout, "cannot get config descriptor %d, %s (%d)\n", nr, strerror(errno), errno);
 		return;
 	}
@@ -1102,6 +1256,10 @@ static void do_config(int fd, unsigned int nr, u_int16_t lang)
 				dump_audiostreaming_interface(fd, p);
 				break;
 			}
+			if (curclass == 1 && cursubclass == 3) {
+				dump_midistreaming_interface(fd, p, lang);
+				break;
+			}
 			printf("  unknown descriptor type:");
 			dump_junk2(p, p[0]);
 			break;
@@ -1111,6 +1269,10 @@ static void do_config(int fd, unsigned int nr, u_int16_t lang)
 				dump_audiostreaming_endpoint(fd, p);
 				break;
 			}
+			if (curclass == 1 && cursubclass == 3) {
+				dump_midistreaming_endpoint(fd, p);
+				break;
+			}
 			printf("  unknown descriptor type:");
 			dump_junk2(p, p[0]);
 			break;
@@ -1145,6 +1307,7 @@ static u_int16_t dump_langids(int fd, int quiet)
 	l = usb_control_msg(fd, USB_DIR_IN, USB_REQ_GET_DESCRIPTOR, USB_DT_STRING << 8, 0, sizeof(b), b);
 
 	if (l < 0) {
+		if (open_mode == O_RDWR)
 		printf("  Language IDs: none (cannot get min. string descriptor; got len=%d, error=%d:%s)\n",
 			l, errno, strerror(errno));
 		return 0;
@@ -1188,18 +1351,22 @@ static void dumpdev(unsigned char *devdesc, int fd, unsigned int flags)
 
 /* ---------------------------------------------------------------------- */
 
-static void dump_one_device(const char *path, unsigned int flags)
+static int dump_one_device(const char *path, unsigned int flags)
 {
         unsigned char buf[USB_DT_DEVICE_SIZE];
 	unsigned int vid, pid;
 	char vendor[128], product[128];
         int fd;
+	int status;
+
+	status=1;
 
 	if ((fd = open(path, O_RDWR)) == -1) {
 		fprintf(stderr, "cannot open %s, %s (%d)\n", path, strerror(errno), errno);
-		return;
+		return(1);
 	}
 	if (usb_control_msg(fd, USB_DIR_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_DEVICE << 8), 0, USB_DT_DEVICE_SIZE, buf) < 0) {
+		if (open_mode == O_RDWR)
 		fprintf(stderr, "cannot get config descriptor, %s (%d)\n", strerror(errno), errno);
 		goto err;
 	}
@@ -1209,13 +1376,15 @@ static void dump_one_device(const char *path, unsigned int flags)
 	get_product_string(product, sizeof(product), vid, pid);
 	printf("Device: ID %04x:%04x %s %s\n", vid, pid, vendor, product);
 	dumpdev(buf, fd, flags);
+	status=0;
  err:
 	close(fd);
+	return(status);
 }
 
 /* ---------------------------------------------------------------------- */
 
-static void list_devices(int bus, int devnum, int vendorid, int productid, unsigned int flags)
+static int list_devices(int bus, int devnum, int vendorid, int productid, unsigned int flags)
 {
         DIR *d, *d2;
         struct dirent *de, *de2;
@@ -1223,11 +1392,14 @@ static void list_devices(int bus, int devnum, int vendorid, int productid, unsig
 	unsigned int vid, pid;
 	char vendor[128], product[128];
         int fd;
+	int status;
+
+	status=1; /* 1 device not found, 0 device found */
 
         d = opendir(procbususb);
         if (!d) {
                 fprintf(stderr, "cannot open %s, %s (%d)\n", procbususb, strerror(errno), errno);
-                return;
+                return(1);
         }
         while ((de = readdir(d))) {
                 if (de->d_name[0] < '0' || de->d_name[0] > '9')
@@ -1243,7 +1415,7 @@ static void list_devices(int bus, int devnum, int vendorid, int productid, unsig
 			if (devnum != -1 && strtoul(de2->d_name, NULL, 0) != devnum)
 				continue;
                         snprintf(buf, sizeof(buf), "%s/%s/%s", procbususb, de->d_name, de2->d_name);
-			if ((fd = open(buf, O_RDWR)) == -1) {
+			if ((fd = open(buf, open_mode)) == -1) {
 				fprintf(stderr, "cannot open %s, %s (%d)\n", buf, strerror(errno), errno);
 				continue;
 			}
@@ -1255,16 +1427,21 @@ static void list_devices(int bus, int devnum, int vendorid, int productid, unsig
 			pid = buf[10] | (buf[11] << 8);
 			if (buf[0] >= USB_DT_DEVICE_SIZE && ((vendorid != -1 && vendorid != vid) || (productid != -1 && productid != pid)))
 				goto err;
+			status=0;
 			get_vendor_string(vendor, sizeof(vendor), vid);
 			get_product_string(product, sizeof(product), vid, pid);
-			printf("\nBus %s Device %s: ID %04x:%04x %s %s\n", de->d_name, de2->d_name, vid, pid, vendor, product);
-			dumpdev(buf, fd, flags);
+			if (verblevel > 0)
+				printf("\n");
+			printf("Bus %s Device %s: ID %04x:%04x %s %s\n", de->d_name, de2->d_name, vid, pid, vendor, product);
+			if (verblevel > 0)
+				dumpdev(buf, fd, flags);
 		err:
 			close(fd);
                 }
                 closedir(d2);
         }
         closedir(d);
+	return(status);
 }
 
 /* ---------------------------------------------------------------------- */
@@ -1285,7 +1462,7 @@ void devtree_devdisconnect(struct usbdevnode *dev)
 {
 }
 
-static void treedump(void)
+static int treedump(void)
 { 
 	int fd;
 	char buf[512];
@@ -1293,11 +1470,12 @@ static void treedump(void)
 	snprintf(buf, sizeof(buf), "%s/devices", procbususb);
 	if ((fd = open(buf, O_RDONLY)) == -1) {
                 fprintf(stderr, "cannot open %s, %s (%d)\n", buf, strerror(errno), errno);
-                return;
+                return(1);
 	}
 	devtree_parsedevfile(fd);
 	close(fd);
         devtree_dump();
+	return(0);
 }
 
 /* ---------------------------------------------------------------------- */
@@ -1306,15 +1484,18 @@ int main(int argc, char *argv[])
 {
         static const struct option long_options[] = {
                 { "version", 0, 0, 'V' },
+		{ "verbose", 0, 0, 'v' },
                 { 0, 0, 0, 0 }
         };
+ 	static const char usage[] = "usage: lsusb [-v] [-x] [-p <procpath>] [-s [<bus>:][<devnum>]] [-d [<vendor>]:[<device>]]\n";
 	int c, err = 0;
 	unsigned int allowctrlmsg = 0, treemode = 0;
 	int bus = -1, devnum = -1, vendor = -1, product = -1;
 	const char *devdump = NULL;
 	char *cp;
+	int status;
 
-	while ((c = getopt_long(argc, argv, "D:vxtp:s:d:V", long_options, NULL)) != EOF) {
+	while ((c = getopt_long(argc, argv, "D:vxtP:p:s:d:V", long_options, NULL)) != EOF) {
 		switch(c) {
                 case 'V':
                         printf("lsusb (" PACKAGE ")  " VERSION "\n");
@@ -1322,6 +1503,7 @@ int main(int argc, char *argv[])
                         
 		case 'v':
 			verblevel++;
+			open_mode = O_RDWR;
 			break;
 
 		case 'x':
@@ -1332,6 +1514,7 @@ int main(int argc, char *argv[])
 			treemode = 1;
 			break;
 
+		case 'P':	/* lspci uses -P; but lsusb 0.9, -p */
 		case 'p':
 			procbususb = optarg;
 			break;
@@ -1372,8 +1555,8 @@ int main(int argc, char *argv[])
 			break;
 		}
 	}
-	if (err) {
-		fprintf(stderr, "usage: lsusb [-v] [-x] [-p <procpath>] [-s [<bus>:][<devnum>]] [-d [<vendor>]:[<device>]]\n");
+	if (err || argc > optind) {
+		fprintf(stderr, usage);
 		exit(1);
 	}
 	if ((err = names_init("./usb.ids")) != 0)
@@ -1381,11 +1564,14 @@ int main(int argc, char *argv[])
 			printf("Error, cannot open USBIDS File \"%s\", %s\n", USBIDS_FILE, strerror(err));
 			exit(1);
 	}
-	if (treemode)
-		treedump();
-	else if (devdump)
-		dump_one_device(devdump, 0);
+	status = 0;
+	if (treemode) {
+		/* treemode requires at least verblevel 1 */
+ 		verblevel += 1 - VERBLEVEL_DEFAULT;
+		status = treedump();
+	} else if (devdump)
+		status = dump_one_device(devdump, 0);
 	else
-		list_devices(bus, devnum, vendor, product, 0);
-	return 0;
+		status = list_devices(bus, devnum, vendor, product, 0);
+	return status;
 }
